from ..i18n import fast_translate
from ..tales import TalesEngine
from ..tales import PythonExpr
from ..tales import StringExpr
from ..tales import NotExpr
from ..tales import ExistsExpr
from ..tales import ImportExpr
from ..tal import RepeatDict

from ..template import BaseTemplate
from ..template import BaseTemplateFile

from .program import MacroProgram as Program

try:
    str = unicode
except NameError:
    pass


class PageTemplate(BaseTemplate):
    """Zope Page Templates.

    To provide a custom translation function, use the ``translate``
    parameter.

    This template class supports encoded input. The default is
    ``'ascii'``, but an alternative encoding may be provided as
    ``encoding``. Note that output is always unicode.

    The ``convert`` parameter specifies a function used to convert a
    non-string value to string. Similarly, ``decode`` is used to
    decode an encoded string to a native string (the default is to use
    coercion, which relies on the program's default encoding --
    usually ``'ascii'``).
    """

    expression_types = {
        'python': PythonExpr,
        'string': StringExpr,
        'not': NotExpr,
        'exists': ExistsExpr,
        'import': ImportExpr,
        }

    default_expression = 'python'

    translate = staticmethod(fast_translate)

    convert = None

    decode = str

    encoding = None

    mode = "xml"

    def __init__(self, *args, **kwargs):
        super(PageTemplate, self).__init__(*args, **kwargs)
        self.__dict__.update(kwargs)
        self.macros = Macros(self)

    @property
    def engine(self):
        return TalesEngine(self.expression_types, self.default_expression)

    def parse(self, body):
        escape = True if self.mode == "xml" else False
        return Program(body, self.mode, self.filename, escape=escape)

    def render(self, translate=None, convert=None, decode=None, **kwargs):
        translate = translate if translate is not None else self.translate
        convert = convert if convert is not None else translate
        decode = decode if decode is not None else self.decode
        encoding = self.encoding

        if encoding is not None:
            def translate(msgid, translate=translate, **kwargs):
                if isinstance(msgid, bytes):
                    msgid = str(msgid, encoding)
                return translate(msgid, **kwargs)

            def decode(inst):
                return str(inst, encoding)

        setdefault = kwargs.setdefault

        setdefault("template", self)
        setdefault("macros", self.macros)
        setdefault("repeat", RepeatDict())
        setdefault("translate", translate)
        setdefault("convert", convert)
        setdefault("decode", decode)
        setdefault("nothing", None)

        return super(PageTemplate, self).render(**kwargs)

    def include(self, stream, econtext, rcontext):
        self.cook_check()
        self._render(stream, econtext, rcontext)


class PageTemplateFile(PageTemplate, BaseTemplateFile):
    pass


class PageTextTemplate(PageTemplate):
    mode = "text"


class PageTextTemplateFile(PageTemplateFile):
    mode = "text"

    def render(self, *args, **kwargs):
        result = super(PageTextTemplateFile, self).render(*args, **kwargs)
        return result.encode(self.encoding or 'utf-8')


class Macro(object):
    __slots__ = "include",

    def __init__(self, render):
        self.include = render


class Macros(object):
    __slots__ = "template",

    def __init__(self, template):
        self.template = template

    def __getitem__(self, name):
        name = name.replace('-', '_')
        self.template.cook_check()

        try:
            function = getattr(self.template, "_render_%s" % name)
        except AttributeError:
            raise KeyError(
                "Macro does not exist: '%s'." % name)

        return Macro(function)

    @property
    def names(self):
        self.template.cook_check()

        result = []
        for name in self.template.__dict__:
            if name.startswith('_render_'):
                result.append(name[8:])
        return result



import traceback

from .tokenize import Token
from .utils import format_kwargs


class TemplateError(Exception):
    def __init__(self, msg, token):
        if not isinstance(token, Token):
            token = Token(token, 0)

        self.msg = msg
        self.token = token
        self.filename = token.filename

    def __str__(self):
        text = "%s\n\n" % self.msg
        text += "   - String:   \"%s\"" % self.token

        if self.filename:
            text += "\n"
            text += "   - Filename: %s" % self.filename

        try:
            line, column = self.token.location
        except AttributeError:
            pass
        else:
            text += "\n"
            text += "   - Location: (%d:%d)" % (line, column)

        return text

    @property
    def offset(self):
        return getattr(self.token, "pos", 0)


class ParseError(TemplateError):
    """An error occurred during parsing.

    Indicates an error on the structural level.
    """


class CompilationError(TemplateError):
    """An error occurred during compilation.

    Indicates a general compilation error.
    """


class TranslationError(TemplateError):
    """An error occurred during translation.

    Indicates a general translation error.
    """


class LanguageError(CompilationError):
    """Language syntax error.

    Indicates a syntactical error due to incorrect usage of the
    template language.
    """


class ExpressionError(LanguageError):
    """An error occurred compiling an expression.

    Indicates a syntactical error in an expression.
    """


class RenderError(TemplateError):
    """An error occurred during rendering."""

    def __init__(self, errors, econtext, rcontext):
        kwargs = rcontext.copy()
        kwargs.update(econtext)

        self._errors = errors
        self._kwargs = kwargs

    def __str__(self):
        # Format keyword arguments; consecutive arguments are indented
        # for readability
        try:
            formatted = format_kwargs(self._kwargs)
        except:
            # the ``pprint.pformat`` method calls the representation
            # method of the arguments; this may fail and since we're
            # already in an exception handler, there's no point in
            # pursuing this further
            formatted = ()

        for index, string in enumerate(formatted[1:]):
            formatted[index + 1] = " " * 15 + string

        out = ["An uncaught exception was raised.", ""]

        for error in self.flatten():
            expression, line, column, filename, exc = error
            out += traceback.format_exception_only(type(exc), exc)
            out.append(" - Expression: \"%s\"" % expression)
            out.append(" - Filename:   %s" % (filename or "<string>"))
            out.append(" - Location:   (%d:%d)" % (line, column))

        out.append(" - Arguments:  %s" % "\n".join(formatted))

        return "\n".join(out)

    def flatten(self):
        for error in self._errors:
            exc = error[-1]
            if isinstance(exc, RenderError):
                for error in exc.flatten():
                    yield error
            else:
                yield error



import os
import re
import sys
import atexit
import hashlib
import shutil
import logging
import tempfile

try:
    import pkg_resources
except ImportError:
    pkg_version = 0
else:
    try:
        distribution = pkg_resources.get_distribution("Chameleon")
    except pkg_resources.DistributionNotFound:
        pkg_version = ""
    else:
        if distribution.has_version():
            pkg_version = distribution.version
        else:
            pkg_version = ""

from .compiler import Compiler
from .loader import ModuleLoader
from .loader import MemoryLoader
from .exc import TemplateError
from .exc import RenderError
from .config import DEBUG_MODE
from .config import AUTO_RELOAD
from .config import EAGER_PARSING
from .config import CACHE_DIRECTORY
from .utils import DebuggingOutputStream
from .utils import Scope

try:
    byte_string = str
    str = unicode
    bytes = byte_string
except NameError:
    def byte_string(string):
        return string.encode('utf-8')

version = sys.version_info[:3]
if version < (3, 0, 0):
    from .py25 import raise_with_traceback
else:
    def raise_with_traceback(exc, tb):
        exc.__traceback__ = tb
        raise exc


XML_PREFIXES = [
    byte_string(prefix) for prefix in(
        "<?xml",                      # ascii, utf-8
        "\xef\xbb\xbf<?xml",          # utf-8 w/ byte order mark
        "\0<\0?\0x\0m\0l",            # utf-16 big endian
        "<\0?\0x\0m\0l\0",            # utf-16 little endian
        "\xfe\xff\0<\0?\0x\0m\0l",    # utf-16 big endian w/ byte order mark
        "\xff\xfe<\0?\0x\0m\0l\0",    # utf-16 little endian w/ byte order mark
        )
    ]

XML_PREFIX_MAX_LENGTH = max(map(len, XML_PREFIXES))

RE_META = re.compile(
    r'\s*<meta\s+http-equiv=["\']?Content-Type["\']?'
    r'\s+content=["\']?([^;]+);\s*charset=([^"\']+)["\']?\s*/?\s*>\s*',
    re.IGNORECASE
    )

log = logging.getLogger('chameleon.template')


def _make_module_loader():
    if CACHE_DIRECTORY:
        path = CACHE_DIRECTORY
    else:
        path = tempfile.mkdtemp()

        @atexit.register
        def cleanup(path=path, shutil=shutil):
            shutil.rmtree(path)

    return ModuleLoader(path)


class BaseTemplate(object):
    """Template base class.

    Input must be unicode (or string on Python 3).
    """

    filename = '<string>'

    _cooked = False

    if DEBUG_MODE or CACHE_DIRECTORY:
        loader = _make_module_loader()
    else:
        loader = MemoryLoader()

    if DEBUG_MODE:
        output_stream_factory = DebuggingOutputStream
    else:
        output_stream_factory = list

    debug = DEBUG_MODE

    def __init__(self, body, **kwargs):
        self.__dict__.update(kwargs)

        self.content_type = self.sniff_type(body)
        self.cook(body)

        if self.__dict__.get('debug') is True:
            self.loader = _make_module_loader()

    def __call__(self, **kwargs):
        return self.render(**kwargs)

    def __repr__(self):
        return "<%s %s>" % (self.__class__.__name__, self.filename)

    @property
    def keep_source(self):
        # By default, we only save the generated source code if we're
        # in debugging mode (to save memory).
        return self.__dict__.get('keep_source', DEBUG_MODE)

    def cook(self, body):
        digest = self._digest(body)
        program = self._cook(body, digest)

        for name, function in program.items():
            if not name.startswith('render'):
                continue

            setattr(self, "_" + name, function)

        self._cooked = True

    def cook_check(self):
        assert self._cooked

    def parse(self, body):
        raise NotImplementedError("Must be implemented by subclass.")

    def render(self, **rcontext):
        stream = self.output_stream_factory()
        econtext = rcontext.pop('econtext', False) or Scope(rcontext)
        self.cook_check()
        try:
            self._render(stream, econtext, rcontext)
        except:
            cls, raised, tb = sys.exc_info()
            errors = rcontext.get('__error__')
            if errors:
                exc = RenderError(errors, econtext, rcontext)
                if not issubclass(cls, RenderError):
                    try:
                        new = type(type(exc).__name__, (RenderError, cls), {})
                        exc.__class__ = new
                    except TypeError:
                        # This may not always be possible; let's raise
                        # the original exception in this case
                        exc = raised
                    else:
                        exc.__dict__.update(raised.__dict__)
                raise_with_traceback(exc, tb)

            raise

        return "".join(stream)

    def sniff_type(self, text):
        """Return 'text/xml' if text appears to be XML, otherwise
        return None.
        """
        if not isinstance(text, bytes):
            text = text.encode('ascii', 'ignore')

        for prefix in XML_PREFIXES:
            if text[:len(prefix)] == prefix:
                return "text/xml"

    def _cook(self, body, digest):
        program = self.parse(body)
        source = self._compile(program)

        if self.keep_source:
            self.source = source

        filename = "%s.py" % digest
        return self.loader.get(filename) or \
               self.loader.build(source, filename)

    def _digest(self, body):
        sha = hashlib.sha1(body.encode('utf-8'))
        sha.update(type(self).__name__.encode('utf-8'))
        sha.update(pkg_version.encode('utf-8'))
        return sha.hexdigest()

    def _compile(self, program):
        compiler = Compiler(self.engine, program)
        return compiler.code


class BaseTemplateFile(BaseTemplate):
    """File-based template base class.

    Relative path names are supported only when a template loader is
    provided as the ``loader`` parameter.
    """

    # The default encoding is only for file content since it is always
    # 8-bit input
    default_encoding = "utf-8"

    # Auto reload is not enabled by default because it's a significant
    # performance hit
    auto_reload = False

    def __init__(self, filename, *args, **kwargs):
        if not os.path.isabs(filename):
            loader = kwargs.get('loader')
            if loader is not None:
                filename = loader.find(filename)

        # Normalize filename
        filename = os.path.abspath(
            os.path.normpath(os.path.expanduser(filename))
            )

        # Make sure file exists
        os.lstat(filename)

        self.filename = filename
        self.auto_reload = kwargs.pop('auto_reload', AUTO_RELOAD)

        self.__dict__.update(kwargs)

        if self.__dict__.get('debug') is True:
            self.loader = _make_module_loader()

        if EAGER_PARSING:
            self.cook_check()

    def cook_check(self):
        if self.auto_reload:
            mtime = self.mtime()

            if mtime != self._v_last_read:
                self._v_last_read = mtime
                self._cooked = False

        if self._cooked is False:
            body = self.read()
            self.cook(body)

    def detect_encoding(self, body):
        # look for an encoding specification in the meta tag
        try:
            body = body.decode('ascii', 'ignore')
        except UnicodeDecodeError:
            pass
        else:
            match = RE_META.search(body)
            if match is not None:
                return match.groups()

        return None, self.default_encoding

    def mtime(self):
        try:
            return os.path.getmtime(self.filename)
        except (IOError, OSError):
            return 0

    def read(self):
        fd = open(self.filename, "rb")
        try:
            body = fd.read(XML_PREFIX_MAX_LENGTH)
        except:
            fd.close()
            raise

        content_type = self.content_type = self.sniff_type(body)

        if content_type == "text/xml":
            body += fd.read()
            fd.close()
            body = body.decode('utf-8')
        else:
            body += fd.read()
            content_type, encoding = self.detect_encoding(body)

            # for HTML, we really want the file read in text mode:
            fd.close()
            fd = open(self.filename, 'rb')
            body = fd.read().decode(encoding or self.default_encoding)
            fd.close()

        return body

    def _get_filename(self):
        return self.__dict__.get('filename')

    def _set_filename(self, filename):
        self.__dict__['filename'] = filename
        self._v_last_read = None

    filename = property(_get_filename, _set_filename)

    body = property(read)

    def _cook(self, body, digest):
        filename = os.path.basename(self.filename)
        base, ext = os.path.splitext(filename)
        name = "%s_%s.py" % (base, digest)
        cooked = self.loader.get(name)

        if cooked is None:
            log.debug('cache miss: %s' % self.filename)
            try:
                nodes = self.parse(body)
                source = self._compile(nodes)

                if DEBUG_MODE:
                    source = "# filename: %s\n#\n%s" % (self.filename, source)

                cooked = self.loader.build(source, name)
            except TemplateError:
                exc = sys.exc_info()[1]
                exc.filename = self.filename
                raise
        else:
            log.debug('cache hit: %s' % self.filename)

        return cooked



# -*- coding: utf-8 -*-

import re
import os
import sys
import shutil
import tempfile

from functools import wraps

try:
    from unittest2 import TestCase
except ImportError:
    from unittest import TestCase

try:
    str = unicode
except:
    pass


class Message(object):
    def __str__(self):
        return "message"


class ImportTestCase(TestCase):
    def test_pagetemplates(self):
        from chameleon import PageTemplate
        from chameleon import PageTemplateFile
        from chameleon import PageTemplateLoader

    def test_pagetexttemplates(self):
        from chameleon import PageTextTemplate
        from chameleon import PageTextTemplateFile


class TemplateFileTestCase(TestCase):
    @property
    def _class(self):
        from chameleon.template import BaseTemplateFile

        class TestTemplateFile(BaseTemplateFile):
            cook_count = 0

            def cook(self, body):
                self.cook_count += 1
                self._cooked = True

        return TestTemplateFile

    def _make_temporary_file(self):
        return tempfile.NamedTemporaryFile(suffix=".py")

    def test_cook_check(self):
        f = self._make_temporary_file()
        template = self._class(f.name)
        template.cook_check()
        self.assertEqual(template.cook_count, 1)

    def test_auto_reload(self):
        f = self._make_temporary_file()

        # set time in past
        os.utime(f.name, (0, 0))

        template = self._class(f.name, auto_reload=True)
        template.cook_check()

        # a second cook check makes no difference
        template.cook_check()
        self.assertEqual(template.cook_count, 1)

        # set current time on file
        os.utime(f.name, None)

        # file is reloaded
        template.cook_check()
        self.assertEqual(template.cook_count, 2)

    def test_relative_is_expanded_to_cwd(self):
        try:
            template = self._class("___does_not_exist___")
        except OSError:
            exc = sys.exc_info()[1]
            self.assertEqual(
                os.getcwd(),
                os.path.dirname(exc.filename)
                )
        else:
            self.fail("Expected OSError.")


class RenderTestCase(TestCase):
    root = os.path.dirname(__file__)

    def find_files(self, ext):
        from ..utils import read_encoded
        inputs = os.path.join(self.root, "inputs")
        outputs = os.path.join(self.root, "outputs")
        for filename in sorted(os.listdir(inputs)):
            name, extension = os.path.splitext(filename)
            if extension != ext:
                continue
            path = os.path.join(inputs, filename)
            f = open(path, 'rb')
            got = read_encoded(f.read())
            f.close()

            # if there's no output file, treat document as static and
            # expect intput equal to output
            import glob
            globbed = tuple(glob.iglob(os.path.join(
                outputs, "%s*%s" % (name.split('-', 1)[0], ext))))

            if not globbed:
                self.fail("Missing output for: %s." % name)

            for output in globbed:
                if not os.path.exists(output):
                    want = got
                else:
                    g = open(output, 'rb')
                    want = read_encoded(g.read())
                    g.close()

                name, ext = os.path.splitext(output)
                basename = os.path.basename(name)
                if '-' in basename:
                    language = basename.split('-')[1]
                else:
                    language = None

                yield path, got, want, language


class ZopePageTemplatesTest(RenderTestCase):
    @property
    def factory(body):
        from ..zpt.template import PageTemplate
        return PageTemplate

    def template(body):
        def decorator(func):
            @wraps(func)
            def wrapper(self):
                template = self.factory(body)
                return func(self, template)

            return wrapper
        return decorator

    def error(body):
        def decorator(func):
            @wraps(func)
            def wrapper(self):
                from ..exc import TemplateError
                try:
                    self.factory(body)
                except TemplateError:
                    exc = sys.exc_info()[1]
                    return func(self, body, exc)
                else:
                    self.fail("Expected exception.")

            return wrapper
        return decorator

    @template("""<span tal:content='str(default)'>Default</span>""")
    def test_default_is_not_a_string(self, template):
        from chameleon.exc import RenderError
        try:
            template()
        except RenderError:
            exc = sys.exc_info()[1]
            self.assertTrue('symbolic value' in str(exc))
        else:
            self.fail("Expected error.")

    @error("""<tal:block replace='bad /// ' />""")
    def test_syntax_error(self, body, exc):
        self.assertTrue(body[exc.offset:].startswith('bad ///'))

    @error("""<tal:dummy attributes=\"dummy 'dummy'\" />""")
    def test_attributes_on_tal_tag_fails(self, body, exc):
        self.assertTrue(body[exc.offset:].startswith('dummy'))

    def test_custom_encoding_for_str_or_bytes(self):
        string = '<div>Тест${text}</div>'
        try:
            string = string.decode('utf-8')
        except AttributeError:
            pass

        template = self.factory(string, encoding="windows-1251")

        text = 'Тест'

        try:
            text = text.decode('utf-8')
        except AttributeError:
            pass

        rendered = template(text=text.encode('windows-1251'))

        self.assertEqual(
            rendered,
            string.replace('${text}', text)
            )

    def test_repr(self):
        from chameleon.zpt.template import PageTemplateFile
        template = PageTemplateFile(
            os.path.join(self.root, 'inputs', 'hello_world.pt')
            )
        self.assertTrue(template.filename in repr(template))

    def test_default_debug_flag(self):
        from chameleon.zpt.template import PageTemplateFile
        from chameleon.config import DEBUG_MODE
        template = PageTemplateFile(
            os.path.join(self.root, 'inputs', 'hello_world.pt'),
            )
        self.assertEqual(template.debug, DEBUG_MODE)
        self.assertTrue('debug' not in template.__dict__)

    def test_debug_flag_on_string(self):
        from chameleon.zpt.template import PageTemplate
        from chameleon.loader import ModuleLoader
        template = PageTemplate(
            open(os.path.join(self.root, 'inputs', 'hello_world.pt')).read(),
            debug=True,
            )
        self.assertTrue(template.debug)
        self.assertTrue(isinstance(template.loader, ModuleLoader))

    def test_debug_flag_on_file(self):
        from chameleon.zpt.template import PageTemplateFile
        from chameleon.loader import ModuleLoader
        template = PageTemplateFile(
            os.path.join(self.root, 'inputs', 'hello_world.pt'),
            debug=True,
            )
        self.assertTrue(template.debug)
        self.assertTrue(isinstance(template.loader, ModuleLoader))


class ZopeTemplatesTestSuite(RenderTestCase):
    def setUp(self):
        self.temp_path = temp_path = tempfile.mkdtemp()

        @self.addCleanup
        def cleanup(path=temp_path):
            shutil.rmtree(path)

    def test_pt_files(self):
        from ..zpt.template import PageTemplate
        from ..zpt.template import PageTemplateFile

        class Literal(object):
            def __init__(self, s):
                self.s = s

            def __html__(self):
                return self.s

            def __str__(self):
                raise RuntimeError(
                    "%r is a literal." % self.s)

        from chameleon.loader import TemplateLoader
        loader = TemplateLoader(os.path.join(self.root, "inputs"))

        self.run_tests(
            ".pt", PageTemplate,
            literal=Literal("<div>Hello world!</div>"),
            content="<div>Hello world!</div>",
            message=Message(),
            load=loader.bind(PageTemplateFile)
            )

    def test_txt_files(self):
        from ..zpt.template import PageTextTemplate
        self.run_tests(".txt", PageTextTemplate)

    def run_tests(self, ext, factory, **kwargs):
        from chameleon.utils import DebuggingOutputStream

        def convert(msgid, **kwargs):
            if isinstance(msgid, Message):
                return translate(msgid, **kwargs)
            return str(msgid)

        def translate(msgid, domain=None, mapping=None, context=None,
                      target_language=None, default=None):
            if default is None:
                default = str(msgid)

            if isinstance(msgid, Message):
                default = "Message"

            if mapping:
                default = re.sub(r'\${([a-z_]+)}', r'%(\1)s', default) % \
                          mapping

            if target_language is None:
                return default

            if domain is None:
                with_domain = ""
            else:
                with_domain = " with domain '%s'" % domain

            stripped = default.rstrip('\n ')
            return "%s ('%s' translation into '%s'%s)%s" % (
                stripped, msgid, target_language, with_domain,
                default[len(stripped):]
                )

        for name, source, want, language in self.find_files(ext):
            if language is not None:
                name += '-' + language

            self.shortDescription = lambda: name
            template = factory(
                source,
                keep_source=True,
                output_stream_factory=DebuggingOutputStream,
                )

            import functools

            params = kwargs.copy()
            params.update({
                'translate': functools.partial(
                    translate, target_language=language),
                'convert': functools.partial(
                    convert, target_language=language),
                })

            try:
                got = template.render(**params)
            except:
                import traceback
                e = traceback.format_exc()
                self.fail("%s\n\n    Example source:\n\n%s" % (e, "\n".join(
                    ["%#03.d%s" % (lineno + 1, line and " " + line or "")
                     for (lineno, line) in
                     enumerate(template.source.split(
                         '\n'))])))

            from doctest import OutputChecker
            checker = OutputChecker()
            if checker.check_output(want, got, 0) is False:
                from doctest import Example
                example = Example(name, want)
                diff = checker.output_difference(
                    example, got, 0)
                self.fail("(%s) - \n%s\n\nCode:\n%s" % (
                    name, diff.rstrip('\n'), template.source.encode('utf-8')))
